# FILE: server.tcl
#
# A generic server that can be instantiated and customized.
#

Oc_Class Net_Server {

    const public common defaultAllowList {}
    const public common defaultDenyList {}
    const public common forbidRemoteConnections 1

  # Protocol to be used on connections established by this server
  # Need good check that doesn't depend on naming conventions
  const public variable protocol 

  # The alias by which this server identifies itself
  const public variable alias = {}

    # Log for logging messages
    public variable log {
        if {[catch {$log Append "Logging messages generated by $this."}]} {
            error "$log is not able to log messages."
        }
        if {[info exists logHandler]} {
            $logHandler delete
        }
        set logHandler [em_eventHandler new $log delete "$this Delete"]
        if {$imadelog} {
            $oldlog Delete
        }
        set oldlog $log
    }
    private variable oldlog = Oc_Nop
    private variable logHandler
    private variable imadelog = 0

  # Boolean - does this server register with a services directory?
  const public variable register = 1

  # List of remote hostname patterns allowed to make connections.
  public variable allow {
        $log Append "$this: New allow: $allow"
  }

  # List of remote hostname patterns forbidden to make connections.
  public variable deny {
        $log Append "$this: New deny: $deny"
  }

  # Track whether the server instance owns, and must clean up
  # the log and protocol instances it uses.
  const private variable imadeprotocol = 0

  # The socket through which this server accepts connections
  private variable socket

  private variable port

  private variable myEventHandlers

  Constructor {args} {
    set allow $defaultAllowList
    set deny $defaultDenyList
    eval $this Configure $args
    if {![info exists log]} {
      # Provide a default log
      em_log New log -title "Server $this Log"
      set imadelog 1
    }
    # Die when $log dies.
    lappend myEventHandlers [em_eventHandler new $log delete "$this Delete"]
    if {![info exists protocol]} {
      # Provide a default protocol
      Net_Protocol New protocol -name "Default protocol"
      set imadeprotocol 1
    }
    # Die when $protocol dies.
    lappend myEventHandlers [em_eventHandler new $protocol delete \
            "$this Delete"]
  }

  # Start the server accepting connections port
  method Start { _port } { 
    if {[info exists socket]} {
      error "Server already started"
    }
    if {$forbidRemoteConnections} {
        set socket [socket -server [list $this VerifyHostAccess] \
            -myaddr localhost $_port]
    } else {
        set socket [socket -server [list $this VerifyHostAccess] \
            $_port]
    }
    set port [lindex [fconfigure $socket -sockname] 2]
    $log Append "Server $this started on port $port..."
    if {$register} {
        $this Register
    }
  }

  method VerifyHostAccess {accsock clientipaddr clientport} {
    set myinfo [fconfigure $accsock -sockname]
    set remoteinfo [fconfigure $accsock -peername]
    set myhostname [lindex $myinfo 1]
    set remotehostname [lindex $remoteinfo 1]
    if {![string match $myhostname $remotehostname]} {
      if {[string match $clientipaddr $remotehostname]} {
        $this RefuseConnection $accsock $remotehostname $clientport
        return
      }
      set allowconnection 0
      foreach pattern $allow {
        if {[string match $pattern $remotehostname]} {
          set allowconnection 1
          break
        }
      }
      foreach pattern $deny {
        if {[string match $pattern $remotehostname]} {
          set allowconnection 0
          break
        }
      }
      if {!$allowconnection} {
        $this RefuseConnection $accsock $remotehostname $clientport
        return
      }
    }
    # This causes all connections created by this server to share its log.
    # When the server created the log, that means when it dies, it will
    # kill its log which will kill all the connections -- against the
    # design which lets connections persist after their server dies.
    Net_Connection New _ -socket $accsock -log $log -protocol $protocol
  }

  private method RefuseConnection {accsock rhost cport} {
    puts $accsock "tell Connection refused"
    close $accsock
    $log Append "Server $this refused connection from $rhost:$cport"
  }

  # Returns true if server is accepting connections, false otherwise.
  method IsListening {} {
    return [info exists socket]
  }

  # Return port we're listening on
  method Port {} {
      if {[info exists port]} {
          return $port
      }
      error "server is not listening"
  }

  # Stop accepting connections
  method Stop {} {
    if {![info exists socket]} {
      error "Server not started"
    }
    $log Append "Stopping server $this ..."
    if {$register} {
        set socketinfo [fconfigure $socket -sockname]
        set port [lindex $socketinfo 2]
        $this Deregister
    }
    unset port
    close $socket
    unset socket
  }

    private variable account

    private method Register {} {
        if {[catch {Net_Account New account -hostname localhost -log $log} msg]} {
            catch {unset account}
            $this RegisterFail $msg {}
            return
        }
        lappend myEventHandlers [em_eventHandler new $account delete \
                [list $this RegisterFail "account died" {}]]
        if {[$account Ready]} {
            $this AccountReady
        } else {
            lappend myEventHandlers [em_eventHandler new $account Ready \
                    [list $this AccountReady]]
        }
    }

    method AccountReady {} {
        set qid [$account Send register [pid] $alias $port]
        Oc_EventHandler New _ $account Reply$qid \
	     [list $this RegisterReply $qid] \
             -groups [list $this $this-$qid]
        Oc_EventHandler New _ $account Timeout$qid \
             [list $this RegisterFail "timeout on registration reply" $qid] \
             -groups [list $this $this-$qid]
    }

    method RegisterReply {qid} {
        Oc_EventHandler DeleteGroup $this-$qid
        set reply [$account Get]
        switch -- [lindex $reply 0] {
            0 {
                $log Append "Server $this registered successfully"
            }
            default {
                $this RegisterFail "registration error" $qid
            }
        }
    }

    method RegisterFail {msg qid} {
        Oc_EventHandler DeleteGroup $this-$qid
        set accountname [Net_Account DefaultAccountName]
        $this Delete
        Oc_Log Log "Unable to register server with\
                localhost:$accountname:\n\t$msg" warning Net_Server
    }

    private method Deregister {} {
        if {![info exists account] || ![$account Ready]} {
            return
        }
        set qid [$account Send deregister [pid] $port]
        Oc_EventHandler New _ $account Reply$qid \
                [list $this DeregisterReply $qid] \
                -groups [list $this $this-$qid]
        Oc_EventHandler New _ $account Timeout$qid \
                [list $this DeregisterFail "timeout on dereg. reply" $qid]] \
                -groups [list $this $this-$qid]
    }

    method DeregisterReply {qid} {
        Oc_EventHandler DeleteGroup $this-$qid
        set reply [$account Get]
        switch -- [lindex $reply 0] {
            0 {
                $log Append "Server $this deregistered successfully"
            }
            default {
                $this DeregisterFail "deregistration error" $qid
            }
        }
    }

    method DeregisterFail {msg qid} {
        Oc_EventHandler DeleteGroup $this-$qid
        set accountname [Net_Account DefaultAccountName]
        # Ought to send up fireworks; this is bad.
        $this Delete
        Oc_Log Log "Unable to deregister server with\
                localhost:$accountname:\n\t$msg" warning Net_Server
    }

  Destructor {
    if {[info exists socket]} {
        # Stop method launches Deregister in background.  Should we wait
        # to confirm it before dying?   Proper usage is to Stop server,
        # then delete it.  More thought needed here...
        $this Stop
    }
    if {[info exists log]} {
        $log Append "Deleting server $this ..."
    }
    # Delete handlers before deleting event-generating members
    foreach handler $myEventHandlers {
        $handler delete
    }
    Oc_EventHandler DeleteGroup $this
    if {$imadelog} {
        $log Delete
    }
    if {$imadeprotocol} {
        $protocol Delete
    }
  }

}
